# File tres/Src/rebin.x
# February 6, 2009
# Adapted by Doug Mink from RVSAO Util/rebin.x

# REBINM rebins a spectrum to a match a specified input wavelength vector

include <mach.h>
include <ctype.h>
include <error.h>
include <imhdr.h>
include <fset.h>
include <math/iminterp.h>
include <math/curfit.h>
include <smw.h>

#rebinning defs	
define RB_NEAREST	1	# nearest neighbor
define RB_LINEAR	2	# linear
define RB_POLY3		3	# 3rd order polynomial
define RB_POLY5		4	# 5th order polynomial
define RB_SPLINE3	5	# cubic spline
define RB_SINC		6	# sinc
define RB_LSINC		7	# look-up table sinc
define RB_DRIZZLE	8	# drizzle
define RB_SUMS		9
define RB_FUNCTIONS	"|nearest|linear|poly3|poly5|spline3|sinc|lsinc|drizzle|sums|"

procedure rebinm (shin, shout, pixin, pixout, debug)

pointer	shin		# Input spectrum header structure
pointer	shout		# Destination spectrum header structure
real	pixin[ARB]	# Input spectrum
real	pixout[ARB]	# Output spectrum (returned)
bool	debug		# True to list process

pointer	invert, sp
char	interp_mode[SZ_LINE]
int	npixin		# Number of pixels in input spectrum
int	npixout		# Number of pixels in output spectrum
double	wl		# Wavelength in Angstroms
double	px,pxo		# Pixel number (1=center of first)
int	mode, user_mode
int	i
int	clgwrd()
double	shdr_lw(), shdr_wl()

begin

# Get rebinning method
	user_mode = clgwrd ("interp_mode", interp_mode, SZ_LINE, RB_FUNCTIONS)
	call cfit_mode (user_mode,mode)

# Interpolate input spectrum to output spectrum
	npixin = SN(shin)
	npixout = SN(shout)
	if (debug) {
	    call printf ("REBINM: %d -> %d pixels\n")
		call pargi (npixin)
		call pargi (npixout)
	    }

# Make room for inverted solution
	call smark (sp)
	call salloc (invert, npixout+1, TY_DOUBLE)

# For SUMS interpolation, compute upper and lower limits of each pixel
	if (mode == RB_SUMS) {
	    do i = 1, npixout+1 {
		pxo = double (i) - 0.5d0
		wl = shdr_lw (shout, pxo)
		px = shdr_wl (shin, wl)
		Memd[invert+i-1] = px
		}
	    call resum (pixin, pixout, Memd[invert], npixin, npixout, debug)
	    if (debug) {
		call printf ("REBINM: %d-element translation vector created\n")
		    call pargi (npixout)
		call flush (STDOUT)
		}
	    }

# For other interpolations, compute pixel centers
	else {
	    do i = 1, npixout {
		pxo = double (i)
		wl = shdr_lw (shout, pxo)
		px = shdr_wl (shin, wl)
		Memd[invert+i-1] = px
#		call printf ("%d: %.4fA -> %.4fA: %.4f\n")
#		    call pargi (i)
#		    call pargd (wl0)
#		    call pargd (wl)
#		    call pargd (px)
		}
	    if (mode == RB_NEAREST)
		call rematch (pixin, pixout, Memd[invert], npixin, npixout)
	    else
		call reinterp (pixin, pixout, Memd[invert], npixin, npixout, mode)
	    }

	call sfree (sp)

end


# CFIT_MODE -- Transform users input mode to CURFIT package mode.

procedure cfit_mode (user_mode, interp_mode)

int	user_mode, interp_mode

begin
	switch (user_mode) {
	    case RB_NEAREST:
		interp_mode = II_NEAREST

	    case RB_LINEAR:
		interp_mode = II_LINEAR

	    case RB_SPLINE3:
		interp_mode = II_SPLINE3

	    case RB_POLY3:
		interp_mode = II_POLY3

	    case RB_POLY5:
		interp_mode = II_POLY5

	    case RB_SINC:
		interp_mode = II_SINC

	    case RB_DRIZZLE:
		interp_mode = II_DRIZZLE

	    case RB_SUMS:
		interp_mode = RB_SUMS
	    }
end


# RESUM -- Rebinning using a partial pixel summation technique to 
#          preserve total flux.

procedure resum (pixin, pixout, invert, npin, npout, debug)

real	pixin[ARB]	# Input spectrum
real	pixout[ARB]	# Output spectrum (returned)
double	invert[ARB]	# Output to input pixel limit map
int 	npin		# Number of pixels in input spectrum
int 	npout		# Number of pixels in output spectrum
bool	debug		# True to list process

int 	ipix
double	x1, x2, xlo, xhi, xmin, xmax

real	pixpart()

begin
	# Initialize
	xmin = 0.5d0
	xmax = double (npin) + 0.5d0
	x2 = invert [1]
	xlo = 0.0
	xhi = 0.0

	do ipix = 1, npout {
	    x1 = x2
	    x2 = invert[ipix + 1]
	    if (x2 < x1) {
		xhi = x1
		xlo = x2
		}
	    else {
		xlo = x1
		xhi = x2
		}

	    # Integrate over the partial pixels from xlo->xhi
	    if (ipix < 30)
		pixout[ipix] = pixpart (pixin, xlo, xhi, debug)
	    else
		pixout[ipix] = pixpart (pixin, xlo, xhi, FALSE)
	    }
end


# PIXPART -- Integrate over partial pixels to obtain total flux
# 	         over specified region.

real procedure pixpart (y, x1, x2, debug)

real	y[ARB]		# Spectrum vector
double	x1		# Starting fractional pixel
double	x2		# Ending fractional pixel
bool	debug		# True to list process

int 	i, i1, i2, npix
double	sum, frac1, frac2, xi1, xi2, v1, v2

begin
	# Remember that pixel centers occur at intgral values
	# so a pixel extends from i-0.5 to i+0.5

	if (debug) {
	    call printf ("%.5f - %.5f: ")
		call pargd (x1)
		call pargd (x2)
	    }
	frac1 = 0.d0
	frac2 = 0.d0

	if (x1 >= x2 || x2 <= x1) {
	    if (debug) {
		call printf ("0.0\n")
		}
	    return (0.d0)
	    }

	# Compute first and last input pixel numbers
	i1 = int (x1 + 0.5d0)
	i2 = int (x2 + 0.5d0)
	sum = 0.d0

	# Flux from part of single pixel
	if (i1 == i2 && i1 > 0) {
	    frac1 = x2 - x1
	    sum = frac1 * y[i1]
	    }

	else {

	# Flux from first pixel
	    if (i1 > 0) {
		xi1 = double (i1) + 0.5d0
		frac1 = xi1 - x1
		v1 = frac1 * y[i1]
		sum = v1
		}

	# Flux from last pixel
	    if (i2 > 0) {
		xi2 = double (i2) - 0.5d0
		frac2 = x2 - xi2
		v2 = frac2 * y[i2]
		sum = sum + v2
		}
	    }

	# Include entire pixels between first and last
	npix = 0
	if (i2 > i1+1) {
	    do i = i1+1, i2-1 {
		sum = sum + y[i]
		npix = npix + 1
		}
	    }
	if (debug) {
	    call printf ("%10.8fx%d=%8g %10.8fx%d=%8g %d = %f\n")
		call pargd (frac1)
		call pargi (i1)
		call pargd (v1)
		call pargd (frac2)
		call pargi (i2)
		call pargd (v2)
		call pargi (npix)
		call pargd (sum)
	    }

	return (sum)
end

# REINTERP -- Rebin the vector by interpolation.

procedure reinterp (pixin, pixout, invert, npin, npout, mode)

real	pixin[ARB]	# Input spectrum
real	pixout[ARB]	# Output spectrum (returned)
double	invert[ARB]	# Output to input pixel map
int 	npin		# Number of pixels in input spectrum
int 	npout		# Number of pixels in output spectrum
int 	mode		# Interpolation mode

int	j
real	xpos

real	arieval()

begin
	do j = 1, npout {
	    xpos = invert[j]
	    if (xpos < 1.0 || xpos > real (npin))
		pixout[j] = 0.0
	    else
		pixout[j] = arieval (xpos, pixin, npin, mode)
	    }
end


# REMATCH -- Rebinning using the closet pixel

procedure rematch (pixin, pixout, invert, npin, npout)

real	pixin[ARB]	# Input spectrum
real	pixout[ARB]	# Output spectrum (returned)
double	invert[ARB]	# Output to input pixel map
int 	npin		# Number of pixels in input spectrum
int 	npout		# Number of pixels in output spectrum

int 	i, i1

begin

	do i = 1, npout {
	    i1 = int (invert[i] + 0.5)
	    if (i1 < 1 || i1 > npin)
		pixout[i] = 0.0
	    else
		pixout[i] = pixin[i1]
	    }
end

# Feb  6 2009	New subroutine based on RVSAO Util/rebin.x
